import requests
from data_type import FieldType

class NocoDB:
    def __init__(self, base_url: str, api_key: str):
        self.api_url = base_url + "/api/v2/meta"
        self.headers = {"xc-token": api_key, "Content-Type": "application/json"}

    def __str__(self) -> str:
        return f"NocoDB API at {self.api_url}"
    
    def __repr__(self) -> str:
        return f"NocoDB(base_url='{self.api_url}', api_key='***')"

    def get_tables(self, base_id: str) -> list["Table"]:
        response = requests.get(
            f"{self.api_url}/bases/{base_id}/tables", headers=self.headers
        )
        if response.status_code == 200:
            tables = [Table(self, **t) for t in response.json()["list"]]
            return tables
        else:
            raise Exception(
                f"Failed to fetch tables with status code: {response.status_code}"
            )


class Table:
    def __init__(self, nocodb: NocoDB | None = None, columns: list["Column"] | None = None, **kwargs) -> None:
        self.nocodb = nocodb
        self.base_id: str = kwargs["base_id"]
        self.table_id: str = kwargs["id"]
        self.title: str = kwargs["title"]
        self.table_name: str = kwargs["table_name"]
        self.columns = columns
        self.meta = kwargs

    def __str__(self) -> str:
        return f"Table '{self.title}' (ID: {self.table_id}, Base: {self.base_id})"

    def __repr__(self) -> str:
        return f"Table(base_id='{self.base_id}', table_id='{self.table_id}', title='{self.title}', columns={self.columns})"

    def get_columns(self) -> list["Column"]:
        if self.nocodb is None:
            raise Exception("NocoDB instance is not provided.")
        response = requests.get(
            f"{self.nocodb.api_url}/tables/{self.table_id}", headers=self.nocodb.headers
        )
        if response.status_code == 200:
            self.columns = [
                Column(self.nocodb, **c)
                for c in response.json()["columns"]
            ]
            return self.columns
        else:
            raise Exception(
                f"Failed to fetch table columns with status code: {response.status_code}"
            )


class Column:
    def __init__(self, nocodb: NocoDB | None = None, **kwargs) -> None:
        self.nocodb = nocodb
        self.title: str = kwargs["title"]
        self.column_name: str = kwargs["column_name"]
        self.column_id: str = kwargs["id"]
        self.table_id: str = kwargs["fk_model_id"]
        self.system = bool(kwargs["system"])
        self.primary_key = bool(kwargs["pk"])
        if isinstance(kwargs["uidt"], FieldType):
            self.data_type = FieldType(kwargs["uidt"])
        else:
            self.data_type = FieldType.from_nocodb(kwargs["uidt"])
        self.meta = kwargs
        self.required = bool(kwargs["rqd"])
        self.options = []
        if self.data_type == FieldType.SINGLE_SELECT:
            self.options = [
                Option(**o) for o in kwargs["colOptions"]["options"]
            ]

    def __str__(self) -> str:
        return f"Column '{self.title}' (ID: {self.column_id}, Table: {self.table_id})"

    def __repr__(self) -> str:
        return f"Column(title='{self.title}', column_name='{self.column_name}', column_id='{self.column_id}', table_id='{self.table_id}', system={self.system}, primary_key={self.primary_key}, data_type='{self.data_type}')"


class Option:
    def __init__(self, title: str, **kwargs) -> None:
        self.title = title
        self.meta = kwargs

    def __str__(self) -> str:
        return f"Options '{self.title}'"
    
    def __repr__(self) -> str:
        return f"Options(name='{self.title}')"


from token import OP
from typing import Any
import requests
from data_type import FieldType


class Teable:
    def __init__(self, base_url: str, api_key: str):
        self.api_url = base_url + "/api"
        self.headers = {
            "Authorization": f"Bearer {api_key}",
            "Accept": "application/json",
        }

    def __str__(self) -> str:
        return f"Teable API at {self.api_url}"

    def __repr__(self) -> str:
        return f"Teable(base_url='{self.api_url}', api_key='***')"

    def get_tables(self, base_id: str) -> list["Table"]:
        response = requests.get(
            f"{self.api_url}/base/{base_id}/table",
            headers=self.headers,
        )
        if response.status_code == 200:
            tables = [Table(**t, teable=self) for t in response.json()]
            return tables
        else:
            raise Exception(
                f"Failed to fetch tables with status code: {response.status_code}, message: {response.text}"
            )

    def create_table(self, base_id: str, table_data: "Table") -> "Table":
        response = requests.post(
            f"{self.api_url}/base/{base_id}/table",
            json=table_data.to_dict(),
            headers=self.headers,
        )
        if response.status_code == 201:
            return response.json()
        else:
            raise Exception(
                f"Failed to create table with status code: {response.status_code}, message: {response.text}"
            )


class Table:
    def __init__(
        self,
        name: str,
        dbTableName: str,
        fields: list["Field"] | None = None,
        teable: Teable | None = None,
        id: str | None = None,
        **kwargs,
    ) -> None:
        self.teable = teable
        self.name = name
        self.dbTableName = dbTableName
        self.fields = fields or []
        self.table_id = id
        self.meta = kwargs

    def __str__(self) -> str:
        return f"Table '{self.name}' (DB Table: {self.dbTableName})"

    def __repr__(self) -> str:
        return f"Table(name='{self.name}', dbTableName='{self.dbTableName}', fields={self.fields})"

    def get_fields(self) -> list["Field"]:
        if self.teable is None:
            raise Exception("Teable instance is not provided.")

        response = requests.get(
            f"{self.teable.api_url}/table/{self.table_id}/field",
            headers=self.teable.headers,
        )
        if response.status_code == 200:
            self.fields = [
                Field(teable=self.teable, field_type=f["type"], **f)
                for f in response.json()
            ]
            return self.fields
        else:
            raise Exception(
                f"Failed to fetch table fields with status code: {response.status_code}, message: {response.text}"
            )
    
    def to_dict(self) -> dict[str, str | list[dict[str, str | dict[str, list[dict[str, str]]]]]]:
        return {
            "name": self.name,
            "dbTableName": self.dbTableName,
            "fields": [field.to_dict() for field in self.fields]
        }

class Field:
    def __init__(
        self,
        name: str,
        dbFieldName: str,
        field_type: str | FieldType,
        options: dict | list["Option"],
        notNull: bool | None = None,
        teable: Teable | None = None,
        **kwargs,
    ) -> None:
        self.teable = teable
        self.name = name
        self.dbFieldName = dbFieldName
        if isinstance(field_type, FieldType):
            self.field_type = field_type
        else:
            self.field_type = FieldType.from_teable(field_type)
        self.notNull = notNull
        self.options = options
        if self.field_type == FieldType.SINGLE_SELECT and isinstance(options, dict):
            self.options = [Option(**o) for o in options["choices"]]
        self.meta = kwargs

    def __str__(self) -> str:
        return f"Field '{self.name}' (DB Field: {self.dbFieldName})"

    def __repr__(self) -> str:
        return f"Field(name='{self.name}', dbFieldName='{self.dbFieldName}', field_type={self.field_type}, options={self.options})"

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "dbFieldName": self.dbFieldName,
            "type": self.field_type.teable_value,
            "options": {"choices": [option.to_dict() for option in self.options]} if self.options else {},
        }

class Option:
    def __init__(self, name: str, **kwargs) -> None:
        self.name = name
        self.meta = kwargs

    def __str__(self) -> str:
        return f"Options '{self.name}'"

    def __repr__(self) -> str:
        return f"Options(name='{self.name}')"

    def to_dict(self) -> dict[str, str]:
        return {
            "name": self.name,
        }



from nocodb import Table as NocoTable, Column as NocoField
from teable import Table as TeaTable, Field as TeaField, Option as TeaOption
from data_type import FieldType as ft


def fieldToTeable(field: NocoField) -> TeaField:
    return TeaField(
        name=field.title,
        dbFieldName=field.column_name,
        field_type=field.data_type,
        notNull=field.required,
        options=[TeaOption(o.title) for o in field.options],
    )


banned_field_type = [
    ft.ID,
    ft.ATTACHMENT,
    ft.MULTIPLE_SELECT,
    ft.FORMULA,
    ft.ROLLUP,
    ft.COUNT,
    ft.LINK,
    ft.BUTTON,
    ft.LINKS,
    ft.AUTO_NUMBER,
    ft.CREATED_TIME,
    ft.CREATED_BY,
    ft.LAST_MODIFIED_BY,
    ft.LAST_MODIFIED_TIME,
    ft.DATE ## Implement this later
]


def tableToTeable(table: NocoTable) -> TeaTable:
    if table.columns is None:
        raise Exception("Table columns are not fetched.")
    fields: list[TeaField] = []
    for f in table.columns:
        if f.data_type not in banned_field_type:
            fields.append(fieldToTeable(f))
    return TeaTable(
        name=table.title,
        dbTableName=table.table_name,
        fields=fields,
    )



from enum import Enum

class FieldType(Enum):
    """Enumeration of field types in NocoDB and Teable."""
    ID = ("ID", "singleLineText")
    SINGLE_LINE_TEXT = ("SingleLineText", "singleLineText")
    LONG_TEXT = ("LongText", "longText")
    USER = ("User", "user")
    ATTACHMENT = ("Attachment", "attachment")
    CHECKBOX = ("Checkbox", "checkbox")
    MULTIPLE_SELECT = ("MultiSelect", "multipleSelect")
    SINGLE_SELECT = ("SingleSelect", "singleSelect")
    DATE = ("Date", "date")
    NUMBER = ("Number", "number")
    DURATION = ("Duration", "duration")
    RATING = ("Rating", "rating")
    FORMULA = ("Formula", "formula")
    ROLLUP = ("Rollup", "rollup")
    COUNT = ("Count", "count")
    LINK = ("LinkToAnotherRecord", "link")
    CREATED_TIME = ("CreatedTime", "createdTime")
    LAST_MODIFIED_TIME = ("LastModifiedTime", "lastModifiedTime")
    CREATED_BY = ("CreatedBy", "createdBy")
    LAST_MODIFIED_BY = ("LastModifiedBy", "lastModifiedBy")
    AUTO_NUMBER = ("AutoNumber", "autoNumber")
    BUTTON = ("Button", "button")
    PHONE_NUMBER = ("PhoneNumber", "singleLineText")
    EMAIL = ("Email", "singleLineText")
    URL = ("URL", "singleLineText")
    LINKS = ("Links", "singleLineText")


    def __init__(self, nocodb_value: str, teable_value: str):
        self.nocodb_value = nocodb_value
        self.teable_value = teable_value

    @classmethod
    def from_nocodb(cls, field_type: str):
        """Convert from NocoDB field type to FieldType."""
        for f in cls:
            if f.nocodb_value == field_type:
                return f
        raise ValueError(f"No matching FieldType for NocoDB value: {field_type}")

    @classmethod
    def from_teable(cls, field_type: str):
        """Convert from Teable field type to FieldType."""
        for f in cls:
            if f.teable_value == field_type:
                return f
        raise ValueError(f"No matching FieldType for Teable value: {field_type}")

    def to_nocodb(self):
        """Get the NocoDB representation of this field type."""
        return self.nocodb_value

    def to_teable(self):
        """Get the Teable representation of this field type."""
        return self.teable_value
